@@ -1,11 +1,11 @@
 import requests
 import json
 
-from dataclasses import asdict
 from typing import List, Union, Optional, IO
 
+from .exceptions import TelegramAPIError
 from .teletypes import (
-    native_type,
+    TelegramStructure,
     Update,
     User,
     Message,
@@ -16,16 +16,17 @@ from .teletypes import (
     Chat,
     ChatMember,
     BotCommand,
-    InlineQueryResult,
     WebhookInfo,
     InputMedia,
     Poll)
 
+from .teletypes.inline_query_result import InlineQueryResult
+
 
 api_url = 'https://api.telegram.org/bot{token}/{method}'
 
 
-def request(token, method, data=dict(), files=dict()):
+def request(token, method, data=dict(), files=dict(), excpect=None):
     """Make a raw api request.
 
     Returns result as dataclass or list of dataclasses
@@ -44,19 +45,21 @@ def request(token, method, data=dict(), files=dict()):
         for key, value in files.items()
         if value is not None
     }
-
-    responce = requests.post(url, data, files=files)
+    json_data = json.dumps(data, ensure_ascii=True)
+    headers = {'Content-type': 'application/json'}
+    responce = requests.post(url, data=json_data, files=files, headers=headers)
 
     if responce.status_code != 200:
-        raise Exception(f'Server returned error: {responce.status_code}')
+        description = \
+            json.loads(responce.text).get('description', 'No description')
+        raise TelegramAPIError(f'[{responce.status_code}] {description}')
 
     data = json.loads(responce.text)['result']
 
     if isinstance(data, dict):
-        return native_type(data)
+        return TelegramStructure.from_dict(data)
     elif isinstance(data, list):
-        for i, item in enumerate(data):
-            data[i] = native_type(item)
+        return TelegramStructure.from_list(data)
     return data
 
 
@@ -72,7 +75,8 @@ def get_updates(token: str,
         data={
             'offset': offset + 1,
             'limit': limit,
-            'allowed_updates': allowed_updates}
+            'allowed_updates': allowed_updates},
+        excpect='update'
         )
     return updates
 
@@ -84,7 +88,8 @@ def get_me(token: str) -> User:
     """
     return request(
         token,
-        'getMe'
+        'getMe',
+        excpect='user'
         )
 
 
@@ -107,13 +112,14 @@ def send_message(token: str,
             'disable_web_page_preview': disable_web_page_preview,
             'disable_notification': disable_notification,
             'reply_to_message_id': reply_to_message_id,
-            'reply_markup': (asdict(reply_markup)
+            'reply_markup': (reply_markup.unfold()
                              if reply_markup is not None else None)
         }
     return request(
         token,
         'sendMessage',
-        data
+        data,
+        excpect='message'
     )
 
 
@@ -136,7 +142,8 @@ def forward_message(token: str,
     return request(
         token,
         'forwardMessage',
-        data
+        data,
+        excpect='message'
     )
 
 
@@ -159,7 +166,7 @@ def send_photo(token: str,
             'parse_mode': parse_mode,
             'disable_notification': disable_notification,
             'reply_to_message_id': reply_to_message_id,
-            'reply_markup': (asdict(reply_markup)
+            'reply_markup': (reply_markup.unfold()
                              if reply_markup is not None else None)
         }
     if not isinstance(photo, str):
@@ -168,12 +175,14 @@ def send_photo(token: str,
                        data,
                        files={
                            'photo': photo
-                       })
+                       },
+                       excpect='message')
     else:
         data['photo'] = photo
         return request(token,
                        'sendPhoto',
-                       data)
+                       data,
+                       excpect='message')
 
 
 def send_audio(token: str,
@@ -202,7 +211,7 @@ def send_audio(token: str,
             'reply_to_message_id': reply_to_message_id,
             'title': title,
             'performer': performer,
-            'reply_markup': (asdict(reply_markup)
+            'reply_markup': (reply_markup.unfold()
                              if reply_markup is not None else None)
         }
     files = dict()
@@ -244,7 +253,7 @@ def send_document(token: str,
             'parse_mode': parse_mode,
             'disable_notification': disable_notification,
             'reply_to_message_id': reply_to_message_id,
-            'reply_markup': (asdict(reply_markup)
+            'reply_markup': (reply_markup.unfold()
                              if reply_markup is not None else None)
         }
 
@@ -296,7 +305,7 @@ def send_video(token: str,
             'supports_streaming': supports_streaming,
             'disable_notification': disable_notification,
             'reply_to_message_id': reply_to_message_id,
-            'reply_markup': (asdict(reply_markup)
+            'reply_markup': (reply_markup.unfold()
                              if reply_markup is not None else None)
         }
 
@@ -346,7 +355,7 @@ def send_animation(token: str,
         'parse_mode': parse_mode,
         'disable_notification': disable_notification,
         'reply_to_message_id': disable_notification,
-        'reply_markup': (asdict(reply_markup)
+        'reply_markup': (reply_markup.unfold()
                          if reply_markup is not None else None)
     }
 
@@ -391,7 +400,7 @@ def send_voice(token: str,
         'parse_mode': parse_mode,
         'disable_notification': disable_notification,
         'reply_to_message_id': disable_notification,
-        'reply_markup': (asdict(reply_markup)
+        'reply_markup': (reply_markup.unfold()
                          if reply_markup is not None else None)
     }
 
@@ -405,7 +414,8 @@ def send_voice(token: str,
         token,
         'sendVoice',
         data,
-        files
+        files,
+        excpect='message'
     )
 
 
@@ -428,7 +438,7 @@ def send_video_note(token: str,
             'length': length,
             'disable_notification': disable_notification,
             'reply_to_message_id': reply_to_message_id,
-            'reply_markup': (asdict(reply_markup)
+            'reply_markup': (reply_markup.unfold()
                              if reply_markup is not None else None)
         }
 
@@ -447,7 +457,8 @@ def send_video_note(token: str,
         token,
         'sendVideoNote',
         data,
-        files
+        files,
+        excpect='message'
     )
 
 
@@ -472,13 +483,14 @@ def send_location(token: str,
         'live_period': live_period,
         'disable_notification': disable_notification,
         'reply_to_message_id': reply_to_message_id,
-        'reply_markup': asdict(reply_markup)
+        'reply_markup': reply_markup.unfold()
     }
 
     return request(
         token,
         'sendLocation',
-        data
+        data,
+        excpect='message'
     )
 
 
@@ -505,13 +517,14 @@ def edit_message_live_location(token: str,
         'longitude': longitude,
         'message_id': message_id,
         'inline_message_id': inline_message_id,
-        'reply_markup': asdict(reply_markup)
+        'reply_markup': reply_markup.unfold()
     }
 
     return request(
         token,
         'editMessageLiveLocation',
-        data
+        data,
+        excpect='message'
     )
 
 
@@ -536,13 +549,14 @@ def stop_message_live_location(token: str,
         'chat_id': chat_id,
         'message_id': message_id,
         'inline_message_id': inline_message_id,
-        'reply_markup': asdict(reply_markup)
+        'reply_markup': reply_markup.unfold()
     }
 
     return request(
         token,
         'stopMessageLiveLocation',
-        data
+        data,
+        excpect='message'
     )
 
 
@@ -573,13 +587,14 @@ def send_venue(token: str,
         'foursquare_type': foursquare_type,
         'disable_notification': disable_notification,
         'reply_to_message_id': reply_to_message_id,
-        'reply_markup': asdict(reply_markup)
+        'reply_markup': reply_markup.unfold()
     }
 
     return request(
         token,
         'sendVenue',
-        data
+        data,
+        excpect='message'
     )
 
 
@@ -605,13 +620,14 @@ def send_contact(token: str,
         'vcard': vcard,
         'disable_notification': disable_notification,
         'reply_to_message_id': reply_to_message_id,
-        'reply_markup': asdict(reply_markup)
+        'reply_markup': reply_markup.unfold()
     }
 
     return request(
         token,
         'sendContact',
-        data
+        data,
+        excpect='message'
     )
 
 
@@ -652,13 +668,14 @@ def send_poll(token: str,
         'is_closed': is_closed,
         'disable_notification': disable_notification,
         'reply_to_message_id': reply_to_message_id,
-        'reply_markup': asdict(reply_markup)
+        'reply_markup': reply_markup.unfold()
     }
 
     return request(
         token,
         'sendPoll',
-        data
+        data,
+        excpect='message'
     )
 
 
@@ -679,13 +696,14 @@ def send_dice(token: str,
         'emoji': emoji,
         'disable_notification': disable_notification,
         'reply_to_message_id': reply_to_message_id,
-        'reply_markup': asdict(reply_markup)
+        'reply_markup': reply_markup.unfold()
     }
 
     return request(
         token,
         'sendDice',
-        data
+        data,
+        excpect='message'
     )
 
 
@@ -790,7 +808,7 @@ def restrict_chat_member(token: str,
     data = {
         'chat_id': chat_id,
         'user_id': user_id,
-        'permissions': asdict(permissions),
+        'permissions': permissions.unfold(),
         'until_date': until_date
     }
 
@@ -876,7 +894,7 @@ def set_chat_permissions(token: str,
         'setChatPermissions',
         {
             'chat_id': chat_id,
-            'permissions': asdict(permissions)
+            'permissions': permissions.unfold()
         }
     )
 
@@ -1162,7 +1180,7 @@ def set_my_commands(token: str,
         token,
         'setMyCommands',
         {
-            'commands': [asdict(i) for i in commands]
+            'commands': [i.unfold() for i in commands]
         }
     )
 
@@ -1173,7 +1191,8 @@ def get_my_commands(token: str) -> List[BotCommand]:
     Returns Array of BotCommand on success.
     """
     return request(
-        token
+        token,
+        'getMyCommands'
     )
 
 
@@ -1195,7 +1214,7 @@ def answer_inline_query(token: str,
         'answerInlineQuery',
         {
             'inline_query_id': inline_query_id,
-            'results': [asdict(i) for i in results],
+            'results': [i.unfold() for i in results],
             'cache_time': cache_time,
             'is_personal': is_personal,
             'next_offset': next_offset,
@@ -1283,7 +1302,7 @@ def edit_message_text(token: str,
             'inline_message_id': inline_message_id,
             'parse_mode': parse_mode,
             'disable_web_page_preview': disable_web_page_preview,
-            'reply_markup': asdict(reply_markup)
+            'reply_markup': reply_markup.unfold()
         }
     )
 
@@ -1311,7 +1330,7 @@ def edit_message_caption(token: str,
             'message_id': message_id,
             'inline_message_id': inline_message_id,
             'parse_mode': parse_mode,
-            'reply_markup': asdict(reply_markup)
+            'reply_markup': reply_markup.unfold()
         }
     )
 
@@ -1337,7 +1356,7 @@ def edit_message_media(token: str,
         'media': media_data,
         'message_id': message_id,
         'inline_message_id': inline_message_id,
-        'reply_markup': asdict(reply_markup)
+        'reply_markup': reply_markup.unfold()
     }
 
     return request(
@@ -1368,7 +1387,7 @@ def edit_message_reply_markup(token: str,
             'chat_id': chat_id,
             'message_id': message_id,
             'inline_message_id': inline_message_id,
-            'reply_markup': asdict(reply_markup)
+            'reply_markup': reply_markup.unfold()
         }
     )
 
@@ -1388,7 +1407,7 @@ def stop_poll(token: str,
         {
             'chat_id': chat_id,
             'message_id': message_id,
-            'reply_markup': asdict(reply_markup)
+            'reply_markup': reply_markup.unfold()
         }
     )
 
@@ -1406,6 +1425,40 @@ def delete_message(token: str,
         'deleteMessage',
         {
             'chat_id': chat_id,
-            'messsage_id': message_id
+            'message_id': message_id
         }
     )
+
+
+def send_media_group(token: str,
+                     chat_id: Union[int, str],
+                     media: List[InputMedia],
+                     disable_notification: Optional[bool] = None,
+                     reply_to_message_id: Optional[bool] = None
+                     ) -> Message:
+    """Use this method to send a group of photos or videos as an album.
+    On success, an array of the sent Messages is returned
+    """
+    assert 10 >= len(media) >= 2, 'Invalid number of media, must be 2-10'
+
+    data = {
+        'chat_id': chat_id,
+        'media': [m.unfold() for m in media],
+        'disable_notification': disable_notification,
+        'reply_to_message_id': reply_to_message_id
+    }
+
+    files = dict()
+    for media_item in media:
+        attach_name, file_path = media_item.file
+        if not all([attach_name, file_path]):
+            continue
+        files[attach_name] = file_path
+
+    return request(
+        token,
+        'sendMediaGroup',
+        data,
+        files,
+        'message'
+    )
